/*
 *  Gideon, a GUI designer for GTK+
 *  Copyright (c) 2005 Maxim Udushlivy
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifndef GIDEON_APPLICATION_H
#define GIDEON_APPLICATION_H

#ifdef GIDEON_SANDBOX
	#define G_DISABLE_DEPRECATED
	#define GDK_DISABLE_DEPRECATED
	#define GDK_PIXBUF_DISABLE_DEPRECATED
	#define PANGO_DISABLE_DEPRECATED
	#define GTK_DISABLE_DEPRECATED
	#define GTKMM_DISABLE_DEPRECATED
#endif

#include <gtk/gtk.h>

#ifdef _WIN32
	#include <gdk/gdkwin32.h>
#endif

#include <assert.h>

#include <gideon.h>
#include <gtkmm/icontheme.h>

#include <vector>
#include <sstream>
#include <iostream>
#include <iomanip>

#ifdef G_LOG_DOMAIN
	#undef G_LOG_DOMAIN
#endif
#define G_LOG_DOMAIN ("Gideon/Application")

#define Check(expr) if(!(expr)) CheckFailed(#expr, __FILE__, __LINE__); else

namespace GideonDesignerApp {

using namespace Gideon;

typedef std::map<std::string, Glib::ustring> Settings;

void CheckFailed(const char * expr, const char * file, unsigned line);
void FatalError(const std::string & message);

std::vector<std::string> Arguments(int argc, char ** argv);
std::vector<Glib::ustring> Split(const Glib::ustring & str, gunichar ch);

GHashTable * SettingsToHashTable(const Settings & settings);
Settings HashTableToSettings(GHashTable * table);

int GetKeyInt(const Settings & settings, const std::string & key);
void SetKeyInt(Settings * settings, const std::string & key, int value);

Glib::ustring GetKeyString(const Settings & settings, const std::string & key);
void SetKeyString(Settings * settings, const std::string & key, const char * const value);
void EnsureKeyString(Settings * settings, const std::string & key, const char * const value);

bool GetKeyBool(const Settings & settings, const std::string & key);
void SetKeyBool(Settings * settings, const std::string & key, bool value);

Gdk::Color GetKeyColor(const Settings & settings, const std::string & key);
void SetKeyColor(Settings * settings, const std::string & key, const Gdk::Color & value);

class ErrorDialog : public Gtk::Window {

public:
	ErrorDialog();

	void onErrorMessage(const Glib::ustring & message);

private:
	Gtk::ScrolledWindow scroll;
	Gtk::TextView textView;

};

class SettingsDialog : public sigc::trackable {

public:
	SettingsDialog();
	~SettingsDialog();

	void setSettings(const Settings & settings);
	Settings getSettings();

	Gtk::Dialog * getWindow() { return window; }

protected:
	void onAdd();
	void onRemove();
	void onMoveUp();
	void onMoveDown();
	void onColorChanged();
	void onCheckChanged();
	void onSelectionChanged();
	void onSwitchPage(GtkNotebookPage *, guint);
	void onDefaults();

	void updateUi(const Settings & settings);
	void updateButtons();
	void updateActions();

	Gtk::TreeIter getPosition();
	void actionCompleted();

private:
	Settings original, defaults;

	Gtk::Dialog * window;	
	Gtk::Button * defaultsButton;
	Gtk::Notebook * widget;
	Gtk::ColorButton * validColor;
	Gtk::ColorButton * wrongColor;
	Gtk::CheckButton * showDebugWidgets;
	Gtk::CheckButton * showHiddenProperties;
	Gtk::CheckButton * suppressMessages;
	Glib::RefPtr<Gtk::ActionGroup> actions;
	Gtk::TreeView * folders;

	Glib::RefPtr<Gtk::ListStore> model;

};

class AppWindow : public Gtk::Window {

public:
	struct State {
		int width;
		int height;
	};

	State getState();
	void setState(const State & state);

};

class Application : public sigc::trackable {

public:
	Application(const std::vector<std::string> & args1, const std::vector<std::string> & args2);
	~Application();

	Gtk::Window * getWindow() { return &window; }

protected:
	struct File {
		File() { header= false; }
		File(const std::string & fn, bool hdr=false) { name= fn; header= hdr; }
		bool empty() { return name.empty(); }
		void clear() { name.clear(); header= false; }
		std::string name;
		bool header;
	};

	void setupHandler(const char * domain);

	static void logHandler(
		const gchar * logDomain,
		GLogLevelFlags logLevel,
		const gchar * message,
		gpointer userData
	);

	void onErrorMessage(
		const Glib::ustring & domain,
		GLogLevelFlags level,
		const Glib::ustring & message
	);

	Glib::RefPtr<Gtk::ActionGroup> createExtendedActions();
	Glib::ustring createExtendedUi();

	void onClear();
	void onClone();
	void onOpen();
	void openFile(const File & filename);
	void onSave();
	void onSaveAs();
	void onQuit();
	void onSettings();

	bool onDeleteEvent(GdkEventAny *);
	void onException();

	bool loadFile(const File & name, Glib::ustring * data);
	bool saveFile(const File & name, const Glib::ustring & data);
	File askFilename(Gtk::FileChooserAction action);
	bool askProceed();
	void updateTitle();

	std::vector<Glib::ustring> currentIconPath();
	void updateIconPath();
	void loadSettings();
	void saveSettings();

	void showMessage(const Glib::ustring & message);

	void designer_extend_ui(const Glib::ustring & ui, const Glib::RefPtr<Gtk::ActionGroup> & actions);
	void designer_apply_settings(const Settings & settings);
	void designer_store_settings(Settings * settings);
	bool designer_can_save();
	bool designer_should_save();
	void designer_clear();
	void designer_load(const Glib::ustring & data);
	Glib::ustring designer_save();
	void designer_saved();
	Glib::RefPtr<Gtk::AccelGroup> designer_get_accel_group();
	Glib::ustring designer_save_header();

private:
	std::vector<Glib::ustring> defaultIconPath;
	Settings settings;
	ErrorDialog errorDialog;

	typedef std::list<std::pair<const char *, guint> > Hids;
	Hids hids;
	
	std::string startupDirectory;
	std::vector<std::string> startupArgs1, startupArgs2;

	File file;

	AppWindow window;
	GideonDesigner * designer;

};

}

#endif
