/*
 *  Gideon, a GUI designer for GTK+
 *  Copyright (c) 2005 Maxim Udushlivy
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#include "headers.h"
#include "forward.h"
#include "designer.h"

#include "gideon_impl.h"
#include "editors.h"
#include "views.h"
#include "palette.h"
#include "canvaswid.h"

namespace Gideon {

// DesignerImpl ----------------------------------------------------------------

DesignerImpl::DesignerImpl() :
	states(50)
{
	SetData(this, "GideonTypeHint", "GideonDesigner");

	actionGroup= createActionGroup();
	uiManager= createUIManager(actionGroup);

	add(vBox);
	vBox.show();

	Gtk::MenuBar * menuBar= dynamic_cast<Gtk::MenuBar *>(uiManager->get_widget("/MainMenubar"));
	vBox.pack_start(*menuBar, Gtk::PACK_SHRINK);

	Gtk::Toolbar * toolbar= dynamic_cast<Gtk::Toolbar *>(uiManager->get_widget("/MainToolbar"));
	toolbar->set_toolbar_style(Gtk::TOOLBAR_ICONS);
	vBox.pack_start(*toolbar, Gtk::PACK_SHRINK);

	statusbar.push("");
	vBox.pack_end(statusbar, Gtk::PACK_SHRINK);
	statusbar.show();
	
	statusbar.pack_end(infoFrame, Gtk::PACK_SHRINK);
	infoFrame.show();

	infoLabel.set_text("");
	infoFrame.add(infoLabel);
	infoLabel.show();

	vBox.pack_start(hPaned, Gtk::PACK_EXPAND_WIDGET);
	hPaned.show();

	hPaned.pack1(vPanedHP, false, true);
	vPanedHP.show();

	hPaned.pack2(hPanedCE, true, true);
	hPanedCE.show();

	paletteWidget.signal_palette_event().connect(sigc::mem_fun(*this, &DesignerImpl::onPaletteEvent));
	vPanedHP.pack2(paletteWidget, true, true);
	paletteWidget.show();

	hPanedCE.pack2(explorerVBox, false, true);
	explorerVBox.show();

	explorerVBox.add(explorerWidget);
	explorerWidget.show();

	vPanedHP.pack1(hierarchyWidget, true, true);
	hierarchyWidget.show();

	hPanedCE.pack1(canvasWidget, true, true);
	canvasWidget.show();

	explorerWidget.getTreeView(ExplorerWidget::epProperties)->setSetRowValues(sigc::mem_fun(*this, &DesignerImpl::setRowValues));
	explorerWidget.getTreeView(ExplorerWidget::epSignals)->setSetRowValues(sigc::mem_fun(*this, &DesignerImpl::setRowValues));
	explorerWidget.getTreeView(ExplorerWidget::epPacking)->setSetRowValues(sigc::mem_fun(*this, &DesignerImpl::setRowValues));

	Gtk::Toolbar * explorerToolbar= dynamic_cast<Gtk::Toolbar *>(uiManager->get_widget("/ExplorerToolbar"));
	explorerToolbar->set_toolbar_style(Gtk::TOOLBAR_ICONS);
	explorerVBox.pack_end(*explorerToolbar, Gtk::PACK_SHRINK);
	explorerToolbar->show();

	hierarchySessionSupplier.signal_session_event().connect(sigc::mem_fun(*this, &DesignerImpl::onHierarchySessionEvent));
	canvasSessionSupplier.signal_session_event().connect(sigc::mem_fun(*this, &DesignerImpl::onCanvasSessionEvent));
	propertiesSessionSupplier.signal_session_event().connect(sigc::mem_fun(*this, &DesignerImpl::onPropertiesSessionEvent));
	propertiesSessionSupplier.signal_after_load().connect(sigc::mem_fun(*this, &DesignerImpl::onAfterLoad));
	signalsSessionSupplier.signal_session_event().connect(sigc::mem_fun(*this, &DesignerImpl::onSignalsSessionEvent));
	signalsSessionSupplier.signal_after_load().connect(sigc::mem_fun(*this, &DesignerImpl::onAfterLoad));
	packingSessionSupplier.signal_session_event().connect(sigc::mem_fun(*this, &DesignerImpl::onPackingSessionEvent));
	packingSessionSupplier.signal_after_load().connect(sigc::mem_fun(*this, &DesignerImpl::onAfterLoad));

	manager.getController()->setContext(this);
	manager.addSupplier(&hierarchySessionSupplier);
	manager.addSupplier(&canvasSessionSupplier);
	manager.addSupplier(&propertiesSessionSupplier);
	manager.addSupplier(&signalsSessionSupplier);
	manager.addSupplier(&packingSessionSupplier);
	manager.acceptSuppliers();

	manager.signal_action_started().connect(sigc::mem_fun(*this, &DesignerImpl::onManagerActionStarted));
	manager.signal_action_completed().connect(sigc::mem_fun(*this, &DesignerImpl::onManagerActionCompleted));
	manager.signal_before_sessions_update().connect(sigc::mem_fun(*this, &DesignerImpl::onBeforeSessionsUpdate));
	manager.signal_after_sessions_update().connect(sigc::mem_fun(*this, &DesignerImpl::onAfterSessionsUpdate));

	hierarchySessionSupplier.signal_popup_menu().connect(sigc::mem_fun(*this, &DesignerImpl::onPopupMenu));

	canvasSessionSupplier.signal_pasting_finished().connect(sigc::mem_fun(*this, &DesignerImpl::onPastingFinished));
	canvasSessionSupplier.signal_status_changed().connect(sigc::mem_fun(*this, &DesignerImpl::onStatusChanged));
	canvasSessionSupplier.signal_popup_menu().connect(sigc::mem_fun(*this, &DesignerImpl::onPopupMenu));

	explorerWidget.signal_selection_changed().connect(sigc::mem_fun(*this, &DesignerImpl::onExplorerSelectionChanged));

	clipboardTarget= "gideon-gui";
	savePoint= -1;

	Palette::get()->signal_leaks().connect(sigc::mem_fun(*this, &DesignerImpl::onLeaks));
	onLeaks();

	updateActions();
}

void DesignerImpl::extend_ui(const Glib::ustring & ui, const PGtkActionGroup & actions) {
	if(extendedActions) {
		uiManager->remove_ui(extendedUi);
		uiManager->remove_action_group(extendedActions);
	}
	extendedActions= actions;
	uiManager->insert_action_group(actions);
	extendedUi= uiManager->add_ui_from_string(ui);
	updateSaveAction();
}

void DesignerImpl::apply_settings(const Settings & set) {
	for(Settings::const_iterator it= set.begin(); it!=set.end(); ++it)
		settings[it->first]= it->second;
	
	if(set.find("hpaned")!=set.end())
		hPaned.set_position(getKeyValue<int>("hpaned"));
	if(set.find("hpaned_ce")!=set.end())
		hPanedCE.set_position(getKeyValue<int>("hpaned_ce"));
	if(set.find("vpaned_hp")!=set.end())
		vPanedHP.set_position(getKeyValue<int>("vpaned_hp"));
	if(set.find("hier_split")!=set.end())
		hierarchyWidget.setSplit(getKeyValue<int>("hier_split"));
	if(set.find("expl_split")!=set.end())
		explorerWidget.setSplit(getKeyValue<int>("expl_split"));
	
	if(set.find("wce_vc")==set.end())
		settings["wce_vc"]= ToString(Gdk::Color("forest green"));
	if(set.find("wce_wc")==set.end())
		settings["wce_wc"]= ToString(Gdk::Color("indian red"));
	
	paletteWidget.create(getKeyValue("show_dw", false));
	manager.applySettings();
}

void DesignerImpl::store_settings(Settings * set) {
	settings["hpaned"]= ToString(hPaned.get_position());
	settings["hpaned_ce"]= ToString(hPanedCE.get_position());
	settings["vpaned_hp"]= ToString(vPanedHP.get_position());
	settings["hier_split"]= ToString(hierarchyWidget.getSplit());
	settings["expl_split"]= ToString(explorerWidget.getSplit());
	manager.storeSettings();
	for(Settings::iterator it= settings.begin(); it!=settings.end(); ++it)
		(*set)[it->first]= it->second;
}

bool DesignerImpl::can_save() {
	return savePoint != manager.getModel()->getPointer();
}

bool DesignerImpl::should_save() {
	return can_save() && !manager.getModel()->emptyHistory();
}

void DesignerImpl::clear() {
	savePoint= -1;
	manager.clear();
	updateSaveAction();
}

void DesignerImpl::load(const Glib::ustring & str) {

	Glib::ustring data;

	if(!str.empty() && str[0]=='"') {
		std::vector<Glib::ustring> lines= Split(str, '\n');
		for(int i=0; i<(int)lines.size(); ++i) {
			int sz= lines[i].size();
			if(sz >= 2)
				data += Glib::strcompress(lines[i].substr(1, sz-2));
			else
				break;
		}
	} else
		data= str;

	savePoint= -1;
	LoadResult result= manager.load(data);
	savePoint= result==lrChanged ? -1 : 0;
	updateSaveAction();
	hierarchyWidget.getEditorWidget()->collapse();
	if(result == lrChanged)
		showChangedWarning();
}

Glib::ustring DesignerImpl::save() {
	Glib::ustring data= manager.save();
	return data;
}

void DesignerImpl::saved() {
	savePoint= manager.getModel()->getPointer();
	updateSaveAction();
}

PGtkAccelGroup DesignerImpl::get_accel_group() {
	return uiManager->get_accel_group();
}

Glib::ustring DesignerImpl::save_header() {
	Glib::ustring data= manager.save();

	std::vector<Glib::ustring> lines= Split(data, '\n');
	data.clear();
	for(int i=0; i<(int)lines.size(); ++i)
		if(!lines[i].empty())
			data += '"' + Glib::strescape(lines[i]+'\n') + "\"\n";
	
	Nodes strings= manager.getModel()->query(FindType("string"));
	for(Nodes::iterator it= strings.begin(); it!=strings.end(); ++it) {
		PNode node= *it;
		const Glib::ustring & value= *node->getValue()->getString();
		if(!value.empty() && manager.getController()->willSave(node).first) {
			Translatable meta(*node->getMeta());
			if(meta.getTranslate()) {
				if(!meta.getComments()->empty())
					data += "\n/*TRANSLATORS: " + *meta.getComments() + "*/";
				Glib::ustring prefix= meta.getPrefix()->empty() ? Glib::ustring() : *meta.getPrefix() + "|";					
				data += "\nGUIXML_(\"" + Glib::strescape(prefix + value)  + "\")\n";
			}
		}
	}
		
	return data;
}

void DesignerImpl::set_embedded(bool embedded) {
	statusbar.set_has_resize_grip(!embedded);
}

void DesignerImpl::sweep() {
	PWidgetCanvasEditor editor= PWidgetCanvasEditor::cast_dynamic(canvasSessionSupplier.getEditor());
	if(editor) {
		PNode node= editor->getSession()->getNode1();
		if(!node->isInactive() && manager.getModel()->raiseMaster(node) != editor->getTopMaster())
			editor->getEditorWidget()->unsetWidget();
	}
}

Gtk::Window * DesignerImpl::getWindow() {
	return dynamic_cast<Gtk::Window *>(get_toplevel());
}

PGtkActionGroup DesignerImpl::createActionGroup() {

	PGtkActionGroup actions= Gtk::ActionGroup::create();

	actions->add(Gtk::Action::create("Edit", "_Edit"));
	actions->add(Gtk::Action::create("Undo", Gtk::Stock::UNDO, "_Undo", "Undo the last action"), Gtk::AccelKey("<control>Z"), sigc::mem_fun(*this, &DesignerImpl::onActionUndo));
	actions->add(Gtk::Action::create("Redo", Gtk::Stock::REDO, "_Redo", "Redo the undone action"), Gtk::AccelKey("<shift><control>Z"), sigc::mem_fun(*this, &DesignerImpl::onActionRedo));
	actions->add(Gtk::Action::create("Cut", Gtk::Stock::CUT, "Cu_t", "Cut the selection"), sigc::mem_fun(*this, &DesignerImpl::onActionCut));
	actions->add(Gtk::Action::create("Copy", Gtk::Stock::COPY, "_Copy", "Copy the selection"), sigc::mem_fun(*this, &DesignerImpl::onActionCopy));
	actions->add(Gtk::Action::create("Paste", Gtk::Stock::PASTE, "_Paste...", "Paste the clipboard"), sigc::mem_fun(*this, &DesignerImpl::onActionPaste));
	actions->add(Gtk::Action::create("PasteToRoot", "Paste to _Root"), sigc::mem_fun(*this, &DesignerImpl::onActionPasteToRoot));
	actions->add(Gtk::Action::create("Delete", Gtk::Stock::DELETE, "_Delete", "Delete the selection"), sigc::mem_fun(*this, &DesignerImpl::onActionDelete));
	
	actions->add(Gtk::Action::create("Help", "_Help"));
	actions->add(Gtk::Action::create("Home", Gtk::Stock::HOME, "_Home"), Gtk::AccelKey("F1"), sigc::mem_fun(*this, &DesignerImpl::onActionHome));
	actions->add(Gtk::Action::create("Forum", Gtk::Stock::HELP, "_Forum"), sigc::mem_fun(*this, &DesignerImpl::onActionForum));
	actions->add(Gtk::Action::create("About", Gtk::Stock::ABOUT, "_About"), sigc::mem_fun(*this, &DesignerImpl::onActionAbout));
	
	actions->add(Gtk::Action::create("Actions", "_Actions"));
	actions->add(Gtk::Action::create("Preview", Gtk::Stock::EXECUTE, "_Preview...", "Show widgets in a preview window"), Gtk::AccelKey("<control>T"), sigc::mem_fun(*this, &DesignerImpl::onActionPreview));
	actions->add(Gtk::Action::create("DefaultSize", Gtk::Stock::ZOOM_100, "D_efault Size", "Restore default widget size"), sigc::mem_fun(*this, &DesignerImpl::onActionDefaultSize));
	actions->add(Gtk::Action::create("ShiftLeft", Gtk::Stock::GO_BACK, "Shift _Left", "Shift the widget left"), sigc::mem_fun(*this, &DesignerImpl::onActionShiftLeft));
	actions->add(Gtk::Action::create("ShiftRight", Gtk::Stock::GO_FORWARD, "Shift _Right", "Shift the widget right"), sigc::mem_fun(*this, &DesignerImpl::onActionShiftRight));
	actions->add(Gtk::Action::create("Translatables", Gtk::Stock::EDIT, "_Strings...", "Manage translatable strings"), sigc::mem_fun(*this, &DesignerImpl::onActionTranslatables));
	
	actions->add(Gtk::Action::create("Default", Gtk::Stock::CLEAR, "_Default", "Set the property to its default value"), sigc::mem_fun(*this, &DesignerImpl::onActionDefault));
	actions->add(Gtk::Action::create("Add", Gtk::Stock::ADD, "_Add", "Add a New Element"), sigc::mem_fun(*this, &DesignerImpl::onActionAdd));
	actions->add(Gtk::Action::create("Remove", Gtk::Stock::REMOVE, "Re_move", "Remove the element"), sigc::mem_fun(*this, &DesignerImpl::onActionRemove));
	actions->add(Gtk::Action::create("MoveUp", Gtk::Stock::GO_UP, "Move _Up", "Move the element up"), sigc::mem_fun(*this, &DesignerImpl::onActionMoveUp));
	actions->add(Gtk::Action::create("MoveDown", Gtk::Stock::GO_DOWN, "Move Do_wn", "Move the element down"), sigc::mem_fun(*this, &DesignerImpl::onActionMoveDown));

	return actions;
}

PGtkUIManager DesignerImpl::createUIManager(PGtkActionGroup actions) {

	PGtkUIManager manager= Gtk::UIManager::create();
	manager->insert_action_group(actions);

	Glib::ustring uistr=
		"<ui>"
		"  <menubar name='MainMenubar'>"
		"    <placeholder name='place1'/>"
		"    <menu action='Edit'>"
		"      <menuitem action='Undo'/>"
		"      <menuitem action='Redo'/>"
		"      <separator/>"
		"      <menuitem action='Cut'/>"
		"      <menuitem action='Copy'/>"
		"      <menuitem action='Paste'/>"
		"      <menuitem action='PasteToRoot'/>"
		"      <menuitem action='Delete'/>"
		"      <placeholder name='place2'/>"
		"    </menu>"
		"    <menu action='Actions'>"
		"      <menuitem action='Preview'/>"
		"      <menuitem action='DefaultSize'/>"
		"      <menuitem action='ShiftLeft'/>"
		"      <menuitem action='ShiftRight'/>"
		"      <menuitem action='Translatables'/>"
		"      <separator/>"
		"      <menuitem action='Default'/>"
		"      <menuitem action='Add'/>"
		"      <menuitem action='Remove'/>"
		"      <menuitem action='MoveUp'/>"
		"      <menuitem action='MoveDown'/>"
		"    </menu>"
		"    <menu action='Help'>"
		"      <menuitem action='Home'/>"
		"      <menuitem action='Forum'/>"
		"      <separator/>"
		"      <menuitem action='About'/>"
		"    </menu>"
		"  </menubar>"
		"  <toolbar  name='MainToolbar'>"
		"    <placeholder name='place3'/>"
		"    <toolitem action='Cut'/>"
		"    <toolitem action='Copy'/>"
		"    <toolitem action='Paste'/>"
		"    <toolitem action='Delete'/>"
		"    <separator/>"
		"    <toolitem action='Undo'/>"
		"    <toolitem action='Redo'/>"
		"    <separator/>"
		"    <toolitem action='Preview'/>"
		"    <toolitem action='DefaultSize'/>"
		"    <toolitem action='ShiftLeft'/>"
		"    <toolitem action='ShiftRight'/>"
		"    <toolitem action='Translatables'/>"
		"  </toolbar>"
		"  <toolbar name='ExplorerToolbar'>"
		"    <toolitem action='Default'/>"
		"    <separator/>"
		"    <toolitem action='Add'/>"
		"    <toolitem action='Remove'/>"
		"    <toolitem action='MoveUp'/>"
		"    <toolitem action='MoveDown'/>"
		"  </toolbar>"
		"  <popup name='MainPopup'>"
		"    <menuitem action='Cut'/>"
		"    <menuitem action='Copy'/>"
		"    <menuitem action='Delete'/>"
		"    <separator/>"
		"    <menuitem action='DefaultSize'/>"
		"    <menuitem action='ShiftLeft'/>"
		"    <menuitem action='ShiftRight'/>"
		"  </popup>"
		"</ui>"
	;

	try {
		manager->add_ui_from_string(uistr);
	} catch(const Glib::Error & x) {
		FatalError("Creating UIManager failed: " + x.what());
	}

	return manager;
}

void DesignerImpl::showChangedWarning() {
	ShowMessage("Warning: one or more properties were changed", this);
}

void DesignerImpl::onActionUndo() {
	manager.undo();
}

void DesignerImpl::onActionRedo() {
	manager.redo();
}

void DesignerImpl::onActionCut() {
	onActionCopy();
	onActionDelete();
}

void DesignerImpl::onActionCopy() {
	clipboardStore= manager.copy();
	std::vector<Gtk::TargetEntry> targets;
	targets.reserve(2);
	targets.push_back(Gtk::TargetEntry(clipboardTarget));
	targets.push_back(Gtk::TargetEntry("UTF8_STRING"));
	Gtk::Clipboard::get()->set(
		targets,
		sigc::mem_fun(*this, &DesignerImpl::onClipboardGet),
		sigc::mem_fun(*this, &DesignerImpl::onClipboardClear)
	);
	Gtk::Clipboard::get()->set_can_store();
	Gtk::Clipboard::get()->store();
}

void DesignerImpl::onActionPaste() {
	Gtk::Clipboard::get()->request_contents(
		clipboardTarget,
		sigc::bind<bool>(sigc::mem_fun(*this, &DesignerImpl::onClipboardReceived), false)
	);
}

void DesignerImpl::onActionPasteToRoot() {
	Gtk::Clipboard::get()->request_contents(
		clipboardTarget,
		sigc::bind<bool>(sigc::mem_fun(*this, &DesignerImpl::onClipboardReceived), true)
	);
}

void DesignerImpl::onActionDelete() {
	manager.deleteSelection();
}

void DesignerImpl::onActionAbout() {

	struct tmp { static void onShowURL(GtkAboutDialog * about, const gchar * link, gpointer data) {
		ShowURL(link, Glib::wrap(about));
	}};

	gtk_about_dialog_set_url_hook(&tmp::onShowURL, NULL, NULL);

	int major= gideon_version()/1000000;
	int minor= gideon_version()/1000 - major*1000;
	int micro= gideon_version() - major*1000000 - minor*1000;
	Glib::ustring version= ToString(major) + "." + ToString(minor) + "." + ToString(micro);

	Gtk::AboutDialog dialog;
	dialog.set_name("Gideon");
	dialog.set_version(version);
	dialog.set_comments("A modern GUI designer for GTK+ toolkit");
	dialog.set_copyright("Copyright \302\251 2005 Maxim Udushlivy");
	dialog.set_website(GideonProjectHome);
	dialog.set_authors(std::vector<std::string>(1, "Maxim Udushlivy <maxim.udushlivy@gmail.com>"));
	dialog.set_license(
		"Gideon is free software; you can redistribute it and/or modify\n"
		"it under the terms of the GNU General Public License as published by\n"
		"the Free Software Foundation; either version 2 of the License, or\n"
		"(at your option) any later version.\n\n"
		"Gideon is distributed in the hope that it will be useful,\n"
		"but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
		"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
		"GNU General Public License for more details.\n\n"
		"You should have received a copy of the GNU General Public License\n"
		"along with Gideon; if not, write to the Free Software\n"
		"Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA"
	);
	dialog.set_transient_for(*getWindow());

	dialog.run();
}

void DesignerImpl::onActionHome() {
	ShowURL(GideonProjectHome, this);
}

void DesignerImpl::onActionForum() {
	ShowURL("http://groups.google.com/group/gideondesigner", this);
}

void DesignerImpl::onActionPreview() {
	gideon_enter_check("DesignerImpl::doPreview()");
	try {
		doPreview();
	} catch(const RuntimeError & x) {
		ShowMessage("Error in preview: " + x.what(), this);
	}
	gideon_leave();
	Palette::get()->checkLeaks();
}

void DesignerImpl::onActionDefaultSize() {
	setDefaultSize();
}

void DesignerImpl::onActionShiftLeft() {
	doShift(-1);
}

void DesignerImpl::onActionShiftRight() {
	doShift(1);
}

void DesignerImpl::onActionTranslatables() {
	TranslatablesDialog dialog;
	dialog.set_transient_for(*getWindow());

	TranslatablesDialog::Strings strings;
	Nodes nodes= manager.getModel()->query(FindType("string"));
	strings.reserve(nodes.size());
	for(Nodes::iterator it= nodes.begin(); it!=nodes.end(); ++it) {
		PNode node= *it;
		bool element= node->getOwner()->getRole()==nrVector;
		Property * property= manager.getController()->findViewProperty(node);
		if(property->hasFlag(sfTranslatable) && property->canEdit()) {
			TranslatablesDialog::String string;
			string.node= node;
			string.path= manager.getModel()->getNodePath(node);
			string.string= *node->getValue()->getString();
			if(!element && property->getDefault())
				string.defval= *property->getDefault()->getString();
			string.meta= Translatable(*node->getMeta());
			strings.push_back(string);
		}
	}
	
	std::sort(strings.begin(), strings.end());

	dialog.setStrings(strings);
	
	if(dialog.run() == Gtk::RESPONSE_OK) {
		manager.begin();
		strings= dialog.getStrings();
		for(int i= 0; i<(int)strings.size(); ++i) {
			manager.getModel()->setScalar(strings[i].node, CAny::createString(strings[i].string));
			manager.getModel()->setMeta(strings[i].node, strings[i].meta.getMeta());
		}
		manager.commit();
	}

}

void DesignerImpl::onActionDefault() {
	doDefault();
}

void DesignerImpl::onActionAdd() {
	doAdd();
}

void DesignerImpl::onActionRemove() {
	doRemove();
}

void DesignerImpl::onActionMoveUp() {
	doMoveUp();
}

void DesignerImpl::onActionMoveDown() {
	doMoveDown();
}

void DesignerImpl::updateActions() {

	actionGroup->get_action("Undo")->set_sensitive(manager.canUndo());
	actionGroup->get_action("Redo")->set_sensitive(manager.canRedo());

	bool selempty= manager.getSelection()->empty();
	actionGroup->get_action("Cut")->set_sensitive(!selempty);
	actionGroup->get_action("Copy")->set_sensitive(!selempty);
	actionGroup->get_action("Delete")->set_sensitive(!selempty);

	actionGroup->get_action("Preview")->set_sensitive(canPreview());
	actionGroup->get_action("DefaultSize")->set_sensitive(canDefaultSize());
	actionGroup->get_action("ShiftLeft")->set_sensitive(canShift());
	actionGroup->get_action("ShiftRight")->set_sensitive(canShift());

	updateSaveAction();
	updateExplorerActions();

}

void DesignerImpl::updateSaveAction() {
	if(extendedActions) {
		PGtkAction action= extendedActions->get_action("Save");
		if(action)
			action->set_sensitive(can_save());
	}
}

void DesignerImpl::updateExplorerActions() {
	ExplorerTest et= testExplorer();
	actionGroup->get_action("Default")->set_sensitive(et & etCanDefault);
	actionGroup->get_action("Add")->set_sensitive(et & etCanAdd);
	actionGroup->get_action("Remove")->set_sensitive(et & etCanRemove);
	actionGroup->get_action("MoveUp")->set_sensitive(et & etCanMoveUp);
	actionGroup->get_action("MoveDown")->set_sensitive(et & etCanMoveDown);
}

void DesignerImpl::onClipboardGet(Gtk::SelectionData & data, guint) {
	std::string target= data.get_target();
	if(target == clipboardTarget)
		data.set(clipboardTarget, clipboardStore);
	else if(target == "UTF8_STRING")
		data.set_text(clipboardStore);
	else if(target == "SAVE_TARGETS")
		;
	else
		ShowMessage("Unexpected clipboard target format", this);
}

void DesignerImpl::onClipboardClear() {
}

void DesignerImpl::onClipboardReceived(const Gtk::SelectionData & data, bool root) {
	std::string target= data.get_target();
	if(target == clipboardTarget) {
 		Glib::ustring clipboard= data.get_data_as_string();
		if(!clipboard.empty()) {
			if(!root)
				paletteWidget.unpush();
			paste(clipboard, root, false);
		}
	} else
		ShowMessage("Unexpected clipboard target format", this);
}

void DesignerImpl::onManagerActionStarted() {
	savedSelection= ToNodeSet(*manager.getSelection());
	if(savedSelection.empty())
		savedState.clear();
	else
		savedState= explorerWidget.getTreeView(ExplorerWidget::epProperties)->saveState();
}

void DesignerImpl::onManagerActionCompleted() {
	Palette::get()->checkLeaks();
	updateActions();
	NodeSet current= ToNodeSet(*manager.getSelection());
	if(current != savedSelection) {
		if(!savedSelection.empty())
			states.set(savedSelection, savedState);
		if(!current.empty()) {
			PObject state= states.get(current);
			if(state)
				explorerWidget.getTreeView(ExplorerWidget::epProperties)->loadState(state);
		}
	}
}

void DesignerImpl::onExplorerSelectionChanged() {
	updateExplorerActions();
}

void DesignerImpl::onBeforeSessionsUpdate() {
	explorerWidget.getTreeView(ExplorerWidget::epProperties)->prepare();
	explorerWidget.getTreeView(ExplorerWidget::epSignals)->prepare();
	explorerWidget.getTreeView(ExplorerWidget::epPacking)->prepare();
}

void DesignerImpl::onAfterSessionsUpdate() {
	explorerWidget.getTreeView(ExplorerWidget::epProperties)->submit();
	explorerWidget.getTreeView(ExplorerWidget::epSignals)->submit();
	explorerWidget.getTreeView(ExplorerWidget::epPacking)->submit();
}

void DesignerImpl::onHierarchySessionEvent(PSession session) {
	if(session->getOperation()==soInsert)
		hierarchyWidget.setEditorWidget(GetPtr(session->getEditor()->getEditorWidget()));
	else if(session->getOperation()==soRemove)
		hierarchyWidget.unsetEditorWidget();
}

void DesignerImpl::onCanvasSessionEvent(PSession session) {
	if(session->getOperation()==soInsert)
		canvasWidget.setEditorWidget(GetPtr(session->getEditor()->getEditorWidget()));
	else if(session->getOperation()==soRemove) {
		onStatusChanged("");
		canvasWidget.unsetEditorWidget();
	}
}

void DesignerImpl::onPropertiesSessionEvent(PSession session) {
	if(session->getOperation()==soInsert || session->getOperation()==soUpdate)
		explorerWidget.getTreeView(ExplorerWidget::epProperties)->update(*session->getPath(), session);
}

void DesignerImpl::onSignalsSessionEvent(PSession session) {
	if(session->getOperation()==soInsert || session->getOperation()==soUpdate)
		explorerWidget.getTreeView(ExplorerWidget::epSignals)->update(*session->getPath(), session);
}

void DesignerImpl::onPackingSessionEvent(PSession session) {
	if(session->getOperation()==soInsert || session->getOperation()==soUpdate)
		explorerWidget.getTreeView(ExplorerWidget::epPacking)->update(*session->getPath(), session);
}

void DesignerImpl::onAfterLoad(PSession session) {
	PPolycell pcvalue= PPolycell::cast_dynamic(session->getEditor()->getEditorWidget());
	PPolycellText pclabel= pcvalue->getElement()->getValue<PolycellText>(0);

	if(session->modelEditable())
		pclabel->changeAttribute(Pango::Attribute::create_attr_style(Pango::STYLE_NORMAL));
	else
		pclabel->changeAttribute(Pango::Attribute::create_attr_style(Pango::STYLE_ITALIC));

	if(session->hasNoUndo() && session->modelEditable())
		pclabel->changeAttribute(Pango::Attribute::create_attr_strikethrough(true));
	else
		pclabel->changeAttribute(Pango::Attribute::create_attr_strikethrough(false));

	if(session->hasHidden())
		pclabel->changeAttribute(Pango::Attribute::create_attr_underline(Pango::UNDERLINE_ERROR));
	else
		pclabel->changeAttribute(Pango::Attribute::create_attr_underline(Pango::UNDERLINE_NONE));

	if(session->hasVectorWithMasterLink() || (session->hasMasterLink() && !session->isElement()))
		pclabel->changeAttribute(Pango::Attribute::create_attr_weight(Pango::WEIGHT_BOLD));
	else
		pclabel->changeAttribute(Pango::Attribute::create_attr_weight(Pango::WEIGHT_NORMAL));

	if(session->isElement())
		pclabel->setText(session->getProperty1()->getLabel(session->getNode1()));
	else
		pclabel->setText((session->hasToSave() ? "*" : "") + session->getName());
}

void DesignerImpl::setRowValues(PPolyelem elem) {
	PSession session= PSession::cast_static(elem->getObject());
	PPolycellText pclabel= NewRefPtr<PolycellText>();
	elem->setValue(0, pclabel);
	PPolycell pcvalue= PPolycell::cast_dynamic(session->getEditor()->getEditorWidget());
	elem->setValue(1, pcvalue);
}

void DesignerImpl::onPaletteEvent(TypeId type, PaletteEvent event) {

	if(event==peCancel) {

		canvasSessionSupplier.stopPasting();

	} else {

		Check(event==peRoot || event==peCanvas);

		Glib::ustring data= Palette::get()->createPasteData(type);
		paste(data, event==peRoot, true);

	}

}

void DesignerImpl::paste(const Glib::ustring & data, bool root, bool normal) {

	if(!root) {

		if(canvasSessionSupplier.startPasting(data, normal))
			return;
		else
			paletteWidget.unpush();

	}

	try {
		manager.pasteToRoot(data, normal);
	} catch(const Glib::Exception & x) {
		ShowMessage("Pasting failed: " + x.what(), this);
	}

}

void DesignerImpl::onPastingFinished() {
	paletteWidget.unpush();
}

void DesignerImpl::onStatusChanged(const Glib::ustring & text) {
	statusbar.pop();
	statusbar.push(text);
}

void DesignerImpl::onPopupMenu() {
	Gtk::Menu * menu= dynamic_cast<Gtk::Menu *>(uiManager->get_widget("/MainPopup"));
	menu->popup(0, GDK_CURRENT_TIME);
}

bool DesignerImpl::canDefaultSize() {
	PNode node= manager.getSingleSelection();
	if(node) {
		Any size;
		if(!manager.getModel()->findMaster(node))
			size= manager.getModel()->getScalar(node, "design-size");
		else
			size= manager.getModel()->getScalar(node, "size-request");
		return size && *size->getPoint() != Point(-1, -1);
	} else
		return false;
}

void DesignerImpl::setDefaultSize() {
	PNode node= manager.getSingleSelection();
	if(node) {
		PNode size;
		if(!manager.getModel()->findMaster(node))
			size= manager.getModel()->find(node, "design-size");
		else
			size= manager.getModel()->find(node, "size-request");
		if(size) {
			manager.begin();
			manager.getModel()->setScalar(size, CAny::createPoint(Point(-1, -1)));
			manager.commit();
		}
	}
}

DesignerImpl::Selected DesignerImpl::getSelected() {
	PSession session= explorerWidget.getSelectedRow();
	if(session) {
		if(!session->isMultiple())
			return std::make_pair(session, session->getNode1());
		else
			return std::make_pair(session, PNode());
	} else if(signalsPageCurrent()) {
		PNode node= manager.getSingleSelection();
		if(node) {
			PNode signals= manager.getModel()->find(node, "signals");
			return std::make_pair(PSession(), signals);
		}
	}
	return Selected();
}

bool DesignerImpl::signalsPageCurrent() {
	return explorerWidget.getCurrentPage() == ExplorerWidget::epSignals;
}

ExplorerTest DesignerImpl::testExplorer() {
	ExplorerTest result= 0;
	Selected selected= getSelected();
	const PSession & session= selected.first;
	const PNode & node= selected.second;

	if(session && session->canDefault())
		result |= etCanDefault;

	if(node) {

		PNode owner= node->getOwner();
		Property * property= manager.getController()->findViewProperty(node);

		if(property->canEdit()) {

			if(node->getRole()==nrVector || owner->getRole()==nrVector)
				result |= etCanAdd;

			if(owner->getRole()==nrVector) {
				result |= etCanRemove;

				int index= FromString<int>(*node->getName());
				int size= owner->getOut()->size();

				if(index > 0)
					result |= etCanMoveUp;

				if(index < size-1)
					result |= etCanMoveDown;

			}

		}

	}

	return result;
}

void DesignerImpl::doDefault() {
	Selected selected= getSelected();
	const PSession & session= selected.first;
	manager.begin();
	if(session->getRole()==srVector)
		session->resetVector();
	else if(session->getRole()==srScalar)
		session->resetScalar();
	else {
		Check(session->getRole()==srObject);
		session->resetObject();
	}
	manager.commit();
}

void DesignerImpl::doAdd() {
	PNode node= getSelected().second;
	PNode vector;
	int index;
	if(node->getRole()==nrVector) {
		vector= node;
		index= 0;
	} else {
		vector= node->getOwner();
		index= FromString<int>(*node->getName()) + 1;
	}
	Property * property= manager.getController()->findViewProperty(vector);
	manager.begin();
	PNode inserted= property->insertElement(index);
	manager.commit();
	if(inserted)
		explorerWidget.setSelectedRow(manager.findSession(inserted));
}

void DesignerImpl::doRemove() {
	PNode node= getSelected().second;
	int index= FromString<int>(*node->getName());
	PNode vector= node->getOwner();
	manager.begin();
	manager.getModel()->vectorRemove(node);
	manager.commit();
	index= std::min(index, (int)vector->getOut()->size() - 1);
	if(explorerWidget.getCurrentPage()!=ExplorerWidget::epSignals || index>=0) {
		PNode nsel= index<0 ? vector : manager.getModel()->find(vector, ToString(index));
		explorerWidget.setSelectedRow(manager.findSession(nsel));
	}
}

void DesignerImpl::doMoveUp() {
	PNode node= getSelected().second;
	manager.begin();
	manager.getModel()->vectorMove(node, -1);
	manager.commit();
	explorerWidget.setSelectedRow(manager.findSession(node));
}

void DesignerImpl::doMoveDown() {
	PNode node= getSelected().second;
	manager.begin();
	manager.getModel()->vectorMove(node, 1);
	manager.commit();
	explorerWidget.setSelectedRow(manager.findSession(node));
}

bool DesignerImpl::canPreview() {
	PNode node= manager.raiseSelectionMaster();
	return node ? Palette::get()->isWidget(node->getType()) : false;
}

void DesignerImpl::doPreview() {
	Gtk::Window hwindow;

	PNode master= manager.raiseSelectionMaster();
	const std::string & name= *master->getName();

	Glib::ustring data= manager.save();
	Gui::PLoader loader= Gui::Loader::create();
	loader->set_preview_mode(true);
	loader->load_from_string(data);
	Gtk::Widget * widget= NULL;
	loader->bind_object(name, &widget);

	Gtk::Window * window= dynamic_cast<Gtk::Window *>(widget);
	if(!window) {
		window= &hwindow;
		window->add(*widget);
		window->set_title("Preview Window");
	} else
		window->set_title("Preview Window [" + window->get_title() + "]");

	gtk_window_set_accept_focus(window->gobj(), true);
	window->set_decorated(true);
	window->set_resizable(true);
	gtk_window_set_focus_on_map(window->gobj(), true);
	window->set_modal(true);
	window->set_type_hint(Gdk::WINDOW_TYPE_HINT_NORMAL);
	window->set_transient_for(*getWindow());
	window->set_position(Gtk::WIN_POS_CENTER_ON_PARENT);
	int W, H;
	window->get_default_size(W, H);
	if(W==-1 || H==-1)
		window->set_default_size(320, 200);

	Gtk::Main::run(*window);
}

void DesignerImpl::onLeaks() {
	int leaks= Palette::get()->getLeaks();
	infoLabel.set_label(leaks==0 ? Glib::ustring() : ("Leaks: " + ToString(leaks)));
}

int DesignerImpl::getShiftCapacity(PNode child) {
	PNode parent= manager.getModel()->findParent(child);
	Any capacity= manager.getModel()->getScalar(parent, "capacity");
	::Gideon::Container * container= manager.getContainer(parent);
	if(dynamic_cast<BoxContainer *>(container)) {
		PNode node= manager.getModel()->findChildNode(child);
		int pack= manager.getModel()->getScalar(node, "pack")->getInt();
		return pack==0 ? capacity->getPoint()->x : capacity->getPoint()->y;
	} else
		return capacity->getInt();
}

PNode DesignerImpl::findShiftNeighbour(PNode child, int index) {
	PNode parent= manager.getModel()->findParent(child);
	::Gideon::Container * container= manager.getContainer(parent);
	int pack= -1;
	if(dynamic_cast<BoxContainer *>(container)) {
		PNode node= manager.getModel()->findChildNode(child);
		pack= manager.getModel()->getScalar(node, "pack")->getInt();
	}
	PNode children= manager.getModel()->find(parent, "children");
	const Nodes & out= *children->getOut();
	for(Nodes::const_iterator it= out.begin(); it!=out.end(); ++it) {
		int cindex= manager.getModel()->find(*it, "index")->getValue()->getInt();
		if(cindex==index) {
			if(pack==-1)
				return manager.getModel()->getWidgetNode(*it);
			else {
				int cpack= manager.getModel()->find(*it, "pack")->getValue()->getInt();
				if(cpack==pack)
					return manager.getModel()->getWidgetNode(*it);
			}
		}
	}
	return PNode();
}

int DesignerImpl::getChildIndex(PNode child) {
	PNode node= manager.getModel()->findChildNode(child);
	return manager.getModel()->getScalar(node, "index")->getInt();
}

void DesignerImpl::setChildIndex(PNode child, int index) {
	PNode node= manager.getModel()->findChildNode(child);
	manager.getModel()->setScalar(node, "index", CAny::createInt(index));
}

bool DesignerImpl::canShift() {
	PNode node= manager.getSingleSelection();
	if(!node)
		return false;
	if(!manager.isIndexChild(node))
		return false;
	int capacity= getShiftCapacity(node);
	return capacity > 1;
}

void DesignerImpl::doShift(int shift) {
	PNode child= manager.getSingleSelection();
	int index= getChildIndex(child);
	int capacity= getShiftCapacity(child);
	int index2= (index+shift+capacity)%capacity;
	PNode child2= findShiftNeighbour(child, index2);
	manager.begin();
	setChildIndex(child, index2);
	if(child2)
		setChildIndex(child2, index);
	manager.commit();
}

}
